# a few convenience functions to query the new crossref API

import urllib.request, urllib.parse, urllib.error, urllib.request, urllib.error, urllib.parse, json, doctest, logging
from . import maxCommon

#def dois(issn):
#    queryStr = ", ".join(queryFields)
#    queryData = {"q" : queryStr, "pages" : "1", "rows" : 1}
#    urlParams = urllib.urlencode(queryData)
#    url = "http://search.crossref.org/dois?" + urlParams
#    # send request
#    jsonStr = urllib2.urlopen(url).read()
#    xrdata = json.loads(jsonStr)
#    # parse result
#    if len(xrdata)==0:
#        logging.debug("Empty cross reply")
#        return None

def lookupDoi(metaInfoDict, repeatCount=2, delaySecs=5):
    """ take author, vol, journal etc from metaInfoDict, query crossref 'links' and return DOI if found

    >>> lookupDoi({"authors":"M. Henrion, D. J. Mortlock, D. J. Hand, and A. Gandy", "title":"A Bayesian approach to star-galaxy classification", "journal":"Monthly Notices of the Royal Astronomical Society", "vol":"414", "issue":"4", "page":"2286", "year":"2011", "printIssn" : ""})
    u'10.1111/j.1365-2966.2010.18055.x'
    """

    # construct url
    mid = metaInfoDict
    logging.debug("Looking up DOI for article %s, %s with crossref links api" % (mid["authors"], mid["title"]))
    freeFormCitFields = [mid["authors"], '"%s"' % mid["title"], mid["journal"],mid["year"], "vol. "+mid["vol"], "no. "+ mid["issue"], "pp. "+mid["page"],  mid["printIssn"]]
    freeFormCitStr = ", ".join(freeFormCitFields)
    logging.debug("crossref.org query %s" % freeFormCitStr)
    url = "https://api.crossref.org/works"

    geturl =  url + "?query=" + urllib.parse.quote(freeFormCitStr.encode('utf-8'))

    # send request
    httpResp = maxCommon.retryHttpRequest(geturl, None, delaySecs=delaySecs, repeatCount=repeatCount)
    if httpResp==None:
        logging.debug("HTTPError while sending crossref request")
        return None

    jsonStr = httpResp.read()
    httpResp.close()
    xrdata = json.loads(jsonStr)

    # parse result
    if len(xrdata)==0:
        logging.debug("Empty cross reply")
        return None

    try:
        items = xrdata["message"]["items"]
    except KeyError:
        logging.debug("Unexpected JSON content from crossref")
        return None
    if len(items) == 0:
        logging.debug("no results in crossref reply")
        return None

    firstRes = items[0]

    logging.debug("Best match from Crossref: %s" % firstRes)
    doi = firstRes["DOI"]
    logging.debug("Got DOI: %s" % doi)
    return doi

if __name__=="__main__":
    import doctest
    doctest.testmod()



# copied from http://lost-theory.org/python/ltchinese/conversion.txt
"""Conversion functions between various Chinese encodings and representations.

The following terms are used to represent the encodings / representation used
in the conversion functions (the samples on the right are for the character
U+4E00 (yi1; "one")):

GB2312 (Kuten/Quwei form): "5027" [used in the "GB2312" field of Unihan.txt]
GB2312 (ISO-2022 form):    "523B" [the "internal representation" of GB code]
EUC-CN:                    "D2BB" [this is the "external encoding" of GB2312-
                                   ISO2022's "internal representation"; also
                                   the form that Ocrat uses]
UTF-8:                     "E4 B8 80" [used in the "UTF-8" field in Unihan.txt]
-------------------------------------------------------------------------------
Unihan UCN:                "U+4E00"   [used by Unicode Inc.]
Unihan NCR (decimal):      "&#19968;" [Numerical Character Reference ...
Unihan NCR (hex):          "&x4E00;"   ... used in XML/HTML/SGML]
-------------------------------------------------------------------------------
internal Python unicode:   u"\\u4e00"  [this is the most useful form!]
internal Python 'utf8':    "\\xe4\\xb8\\x80"
internal Python 'gb2312':  "\\xd2\\xbb"
internal Python 'euc-cn':  "\\xd2\\xbb"
internal Python 'gb18030': "\\xd2\\xbb"

See these resources for more information:
 * Wikipedia "Extended_Unix_Code" article
   * "EUC-CN is the usual way to use the GB2312 standard for simplified Chinese
     characters ... the ISO-2022 form of GB2312 is not normally used"
 * Wikipedia "HZ_(encoding)" article (the example conversion)
 * Wikipedia "Numeric_character_reference" article
 * Unihan (look for "Encoding forms", "Mappings to Major Standards")
   * e.g. http://www.unicode.org/cgi-bin/GetUnihanData.pl?codepoint=4E00
"""

import re

hexd = lambda s: hex(s)[2:] #only the hex digits (strip '0x' at the beginning)

def kuten_to_gb2312(kuten):
    """Convert GB kuten / quwei form (94 zones * 94 points) to GB2312-1980 / ISO-2022-CN hex (internal representation)"""
    zone, point = int(kuten[:2]), int(kuten[2:])
    hi, lo = hexd(zone + 0x20), hexd(point + 0x20)
    return "%s%s" % (hi, lo)

def gb2312_to_euc(gb2312hex):
    """Convert GB2312-1980 hex (internal representation) to EUC-CN hex (the "external encoding")"""
    hi,lo = int(gb2312hex[:2], 16), int(gb2312hex[2:], 16)
    hi,lo = hexd(hi + 0x80), hexd(lo + 0x80)
    return "%s%s" % (hi, lo)

def euc_to_utf8(euchex):
    """Convert EUC hex (e.g. "d2bb") to UTF8 hex (e.g. "e4 b8 80")"""
    utf8 = euc_to_python(euchex).encode("utf-8")
    utf8 = repr(utf8)[1:-1].replace("\\x", " ").strip()
    return utf8

#TODO: make these conversions go both ways (i.e. gb2312_to_kuten, euc_to_gb2312, utf8_to_euc)

#TODO: fill in the conversions? (i.e. create kuten_to_utf8, so you don't have to compose kuten_to_gb2312 and gb2312_to_euc by hand)
#TODO: might even want to create a more generic system so that you can just pick the encoding types and dispatch the right method

#TODO: more encoding formats? (HZ at least, it's easy) ; (actually HZ is an included in the codecs.)

## convert to internal Python unicode / string objects ########################

def ucn_to_python(ucn):
    """Convert a Unicode Universal Character Number (e.g. "U+4E00" or "4E00") to Python unicode (u'\\u4e00')"""
    ucn = ucn.strip("U+")
    if len(ucn) > 4:
        return eval("u'\\U%08x'" % int(ucn, 16)) #this would be a security hole if we didn't use `int` to make the input safe
    else:
        return eval("u'\\u%s'" % ucn) #4 characters isn't enough room to do damage with the eval
        #TODO: this dies on decimal input (e.g. ucn_to_python("100")

def euc_to_python(hexstr):
    """Convert a EUC-CN (GB2312) hex to a Python unicode string"""
    hi = hexstr[0:2]
    lo = hexstr[2:4]
    gb_enc = eval("'\\x%s\\x%s'" % (hi,lo)) #hi and lo are only 2 characters long, no risk with eval-ing them
    return gb_enc.decode("gb2312")

def ncr_to_python(ncr):
    """Convert Unicode Numerical Character Reference (e.g. "19968", "&#19968;", or "&#x4E00;") to native Python Unicode (u'\\u4e00')"""
    ncr = ncr.lower()
    if ncr.startswith("&#x"):
        #hex NCR
        ncr = ncr.strip("&#x;")
    elif ncr.startswith("&#"):
        #decimal NCR
        ncr = ncr.strip("&#;")
        ncr = hexd(int(ncr))
    else:
        #assume it's a decimal NCR not in the XML character entity ref. format
        ncr = hexd(int(ncr))
    return ucn_to_python(ncr)

## convert from internal Python unicode / string objects ######################

def python_to_ucn(uni_char):
    """Converts a one character Python unicode string (e.g. u'\\u4e00') to the corresponding Unicode UCN ('U+4E00')"""
    ucn = repr(uni_char)[4:-1]
    if len(ucn) > 4:
       ucn = ucn.lstrip("0") #get rid of the zeroes that Python uses to pad 32 byte UCNs
    return "U+%s" % ucn

def python_to_euc(uni_char):
    """Converts a one character Python unicode string (e.g. u'\\u4e00') to the corresponding EUC hex ('d2bb')"""
    return repr(uni_char.encode("gb2312"))[1:-1].replace("\\x", "")

def python_to_ncr(uni_char, **options):
    """Converts a one character Python unicode string (e.g. u'\\u4e00') to the corresponding Unicode NCR ('&x4E00;')
    Change the output format by passing the following parameters:
     * no parameters - default behavior: hex=True, xml=True
     * decimal=True - output the decimal value instead of hex
     * hex=False - (same as decimal=True)
     * xml=False - just display the decimal or hex value, i.e. strip off the '&#', '&x', and ';'"""
    hexflag, decflag, xmlflag = list(map(options.get, ["hex", "decimal", "xml"], [True, False, True]))
    if decflag:
        out = int(repr(uni_char)[4:-1], 16)
        if xmlflag:
            out = "&#%s;" % out
    elif hexflag:
        out = repr(uni_char)[4:-1]
        if xmlflag:
            out = "&#x%s;" % out
    return out

## handle multiple characters #################################################

def string_to_ncr(uni_string, **options):
    """Converts a Python unicode string (e.g. u'p\\u012bn y\\u012bn') to the corresponding Unicode NCRs.
    See `python_to_ncr` for formatting options."""
    for char in uni_string:
        if ord(char) > 128:
            uni_string = uni_string.replace(char, python_to_ncr(char, options=options))
    return uni_string

def ncrstring_to_python(ncr_string):
    """Convert a string of Unicode NCRs (e.g. "&#19968;&#x4E00;") to native Python Unicode (u'\\u4e00\\u4e00')"""
    res = re.findall("&#[x0-9a-fA-F]*?;", ncr_string)
    for r in res:
        ncr_string = ncr_string.replace(r, ncr_to_python(r))
    return ncr_string



# general routines to make http connections

from os.path import *
import os, logging, telnetlib, urllib.parse, time, tempfile, urllib.request, urllib.parse, urllib.error, random

import chardet
from . import pubGeneric

httpTimeout = 30
httpUserAgent = 'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:7.0.1) Gecko/20100101 Firefox/7.0.120'
TEMPDIR = "/scratch/tmp"

WGETOPTIONS = " --no-check-certificate --tries=3 --random-wait --waitretry=%d --connect-timeout=%d --dns-timeout=%d --read-timeout=%d --ignore-length --user-agent='%s'" % (httpTimeout, httpTimeout, httpTimeout, httpTimeout, httpUserAgent)

def recodeToUtf8(data):
    " use chardet to find out codepage and recode to utf8"
    encoding = chardet.detect(data)['encoding']
    logging.log(5, "encoding should be %s" % encoding)
    try:
        data = data.decode(encoding).encode("utf8")
    except UnicodeDecodeError:
        logging.warn("Error when decoding as %s" % encoding)
        data = data
    return data

lastCallSec = {}

def wait(delaySec, host="default"):
    " make sure that delaySec seconds have passed between two requests to host "
    global lastCallSec
    delaySec = float(delaySec)
    nowSec = time.time()
    sinceLastCallSec = nowSec - lastCallSec.get(host, 0)
    #logging.debug("sinceLastCall %f" % float(sinceLastCallSec))
    if sinceLastCallSec > 0.1 and sinceLastCallSec < delaySec :
        waitSec = delaySec - sinceLastCallSec
        logging.debug("Waiting for %f seconds" % waitSec)
        time.sleep(waitSec)

    lastCallSec[host] = time.time()

defaultDelay = 20
wgetCache = {}

crawlDelays = {}

def delayedWget(url, forceDelaySecs=None):
    " download with wget and make sure that delaySecs (global var) secs have passed between two calls "
    global wgetCache
    if url in wgetCache:
        logging.log(5, "Using cached wget results")
        return wgetCache[url]

    if forceDelaySecs==None:
        host = urllib.parse.urlsplit(url)[1]
        logging.debug("Looking up delay time for host %s" % host)
        if host in crawlDelays:
            delaySecs = crawlDelays.get(host, defaultDelay)
            logging.debug("Delay time for host %s configured as %d seconds" % (host, delaySecs))
        else:
            logging.debug("Delay time for host %s not configured" % (host))
            delaySecs = defaultDelay
    else:
        delaySecs = forceDelaySecs
        host = "noHost"

    wait(delaySecs, host)
    page = runWget(url)
    return page

def parseWgetLog(logFile, origUrl):
    " parse a wget logfile and return final URL (after redirects) and mimetype as tuple"
    #   Content-Type: text/html; charset=utf-8
    lines = logFile.readlines()
    logging.log(5, "Wget logfile: %s" % " / ".join(lines))
    mimeType, url, charset = None, None, "utf8"
    lastUrl = None
    for l in lines:
        l = l.strip()
        if l.lower().startswith("content-type:"):
            logging.log(5, "wget mime type line: %s" % l)
            mimeParts = l.strip("\n").split()
            if len(mimeParts)>1:
                mimeType = mimeParts[1].strip(";")
            if len(mimeParts)>2:
                charset = mimeParts[2].split("=")[1]

        elif l.lower().startswith("location:"):
            logging.log(5, "wget location line: %s" % l)
            url = l.split(": ")[1].split(" ")[0]
            scheme, netloc = urllib.parse.urlsplit(url)[0:2]
            if netloc=="":
                #assert(lastUrl!=None)
                if lastUrl==None:
                    lastUrl = origUrl
                logging.log(5, "joined parts are %s, %s" % (lastUrl, url))
                url = urllib.parse.urljoin(lastUrl, url)
                logging.log(5, "URL did not contain server, using previous server, corrected URL is %s" % url)
            else:
                lastUrl = url

    logging.log(5, "parsed transfer log files:  URL=%s, mimetype=%s, charset=%s" % (url, mimeType, charset))
    return mimeType, url, charset

def setTorExitNode(host):
    " establish a control connection to tor on localhost:9051 and set the current exit node "
    logging.debug("Setting tor exit node to %s" % host)
    tn = telnetlib.Telnet("127.0.0.1", 9051)

    tn.write('authenticate ""\n')
    reIdx, match, text = tn.expect(["250 OK"], 3)
    #logging.debug("Received %s from tor" % text)
    text = tn.read_eager()
    #logging.debug("Received %s from tor" % text)
    assert(reIdx==0)

    tn.write('setconf ExitNodes=%s\n' % host)
    reIdx, match, text = tn.expect(["250 OK"], 3)
    #logging.debug("Received %s from tor" % text)
    assert(reIdx==0)
    tn.read_eager()
    tn.close()

torNodes = None
currentTorNode = 0


def runWget(url, useTor=None, tmpDir='/tmp'):
    """ download url with wget and return dict with keys url, mimeType, charset, data

    tor support requires a running tor on localhost with an activated control connection
    port and polipo listening on port 8118 and connected to tor.

    """

    global torNodes
    global currentTorNode

    if useTor:
        # download the current exit node list every 60 minutes
        torFname = join(tmpDir, "Tor_ip_list_EXIT.csv")
        if (isfile(torFname) and os.path.getmtime(torFname)-time.time() > 3600) or \
            not isfile(torFname):
            exitNodeUrl = "http://torstatus.blutmagie.de/ip_list_exit.php/Tor_ip_list_EXIT.csv"
            logging.info("Downloading current tor exit node lists from %s" % exitNodeUrl)
            exitNodeData = urllib.request.urlopen(exitNodeUrl).read()
            open(torFname, "w").write(exitNodeData)
            logging.info("wrote new tor list to %s" % torFname)

        if torNodes==None:
            torNodes = []
            for line in open(torFname):
                ip = line.strip()
                torNodes.append(ip)
            random.shuffle(torNodes)

        if currentTorNode > len(torNodes):
            currentTorNode = 0

        os.environ["http_proxy"] = "http://127.0.0.1:8118"
        setTorExitNode(torNodes[currentTorNode])
        currentTorNode += 1

    # check if file is already in cache
    global wgetCache
    if url in wgetCache:
        logging.log(5, "Using cached wget results")
        return wgetCache[url]

    logging.debug("Downloading %s" % url)
    url = url.replace("'", "")

    # run wget command
    tmpFile = tempfile.NamedTemporaryFile(dir = TEMPDIR, prefix="wgetTemp", suffix=".data")
    cmd = "wget '%s' -O %s --server-response" % (url, tmpFile.name)
    cmd += WGETOPTIONS
    logFile = tempfile.NamedTemporaryFile(dir = TEMPDIR, \
        prefix="pubGetPmid-Wget-", suffix=".log")
    cmd += " -o %s " % logFile.name
    stdout, stderr, ret = pubGeneric.runCommandTimeout(cmd, timeout=httpTimeout)
    if ret!=0:
        raise Exception("wgetRetNonNull\t"+url.decode("utf8"))

    # parse wget log
    mimeType, redirectUrl, charset = parseWgetLog(logFile, url)
    if mimeType==None:
        raise Exception("No mimetype found in http reply\t"+url)

    if redirectUrl!=None:
        finalUrl = redirectUrl
    else:
        finalUrl = url

    data = tmpFile.read()
    logging.log(5, "Download OK, size %d bytes" % len(data))
    if len(data)==0:
        raise Exception("empty http reply from %s" % url)

    if mimeType in ["text/plain", "application/xml", "text/csv"]:
        logging.log(5, "Trying to guess encoding of text file with type %s" % mimeType)
        data = recodeToUtf8(data)

    ret = {}
    ret["url"] = finalUrl
    ret["mimeType"] = mimeType
    ret["encoding"] = charset
    ret["data"] = data

    wgetCache[finalUrl] = ret
    wgetCache[url] = ret

    if useTor:
        del os.environ["http_proxy"]

    return ret

def httpStartsWith(urlPrefix, url):
    """check if url starts with urlPrefix, which should be an http: prefix.  This will also check
    with https:"""
    if url.startswith(urlPrefix):
        return True
    return url.startswith(urlPrefix.replace("http:", "https:", 1))



import logging, os, sys, tempfile, csv, collections, types, codecs, gzip, \
    os.path, re, glob, time, urllib.request, urllib.error, urllib.parse, doctest, http.client, socket, io, subprocess, shutil, atexit
from types import *
from os.path import isfile, isdir, getsize, abspath, join, realpath, dirname
from collections import defaultdict

# global flag to suppress all removal of temporary files, only useful for debugging
keepTemp = False

tempPaths = []

def delTemp():
    if keepTemp:
        logging.info("Not deleting temporary files: %s" % ",".join(tempPaths))
        return

    for tmpPath in tempPaths:
        if tmpPath!=None and isdir(tmpPath):
            logging.debug("Deleting dir+subdirs %s" % tmpPath)
            shutil.rmtree(tmpPath)
        elif tmpPath!=None and isfile(tmpPath):
            logging.debug("Deleting file %s" % tmpPath)
            os.remove(tmpPath)
        else:
            # has already been deleted
            pass

def delOnExit(path):
    "make sure that path or file gets deleted upon program exit"
    global tempPaths
    atexit.register(delTemp)
    tempPaths.append(path)

def ignoreOnExit(path):
    " undo delOnExit "
    global tempPaths
    tempPaths.remove(path)

def which(program):
    import os
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

def errAbort(text):
    raise Exception(text)

def mustExistDir(path, makeDir=False):
    if not os.path.isdir(path):
        if makeDir:
            logging.info("Creating directory %s" % path)
            os.makedirs(path)
        else:
            msg = "Directory %s does not exist " % path
            logging.error(msg)
            raise Exception(msg)

def mustExist(path):
    if not (os.path.isdir(path) or os.path.isfile(path)):
        logging.error("%s is not a directory or file" % path)
        sys.exit(1)

def mustNotExist(path):
    if (os.path.isdir(path) or os.path.isfile(path)):
        logging.error("%s already exists" % path)
        sys.exit(1)

def mustNotBeEmptyDir(path):
    mustExist(path)
    fileList = os.listdir(path)
    if len(fileList)==0:
        logging.error("dir %s does not contain any files" % path)
        sys.exit(1)

def makeOrCleanDir(path):
    " empty directory if exists or make it "
    assert(not isfile(path))
    logging.debug("Making/cleaning dir %s" % path)
    if isdir(path):
       shutil.rmtree(path)
    os.makedirs(path)

def deleteFiles(fnames):
    " remove all files "
    if len(fnames)==0:
        logging.debug("Not deleting any files")
        return

    logging.debug("Deleting %d files (%s,...)" % (len(fnames), fnames[0]))
    for fn in fnames:
        os.remove(fn)

def mustBeEmptyDir(path, makeDir=False):
    " exit if path does not exist or it not empty. do an mkdir if makeDir==True "
    if type(path)==list:
        for i in path:
            notEmptyDirs = []
            notExistDirs = []
            if not os.path.isdir(i):
                if makeDir:
                    os.makedirs(i)
                else:
                    notExistDirs.append(i)
            else:
                if len(os.listdir(i))!=0:
                    notEmptyDirs.append(i)
        text = ""
        if len(notEmptyDirs)!=0:
            text += "Directories %s are not empty. " % " ".join(notEmptyDirs)
        if len(notExistDirs)!=0:
            text += "Directories %s do not exist. " % " ".join(notExistDirs)
        if text!="":
            raise Exception(text)
    else:
        if not os.path.isdir(path):
            if not makeDir:
                raise Exception("Directory %s does not exist" % path)
            else:
                logging.info("Creating directory %s" % path)
                os.makedirs(path)
        else:
            if len(os.listdir(path))!=0:
                raise Exception("Directory %s is not empty" % path)

def makeTempFile(tmpDir=None, prefix="tmp", ext=""):
    """ return a REAL temporary file object
    the user is responsible for deleting it!
    """
    fd, filename = tempfile.mkstemp(suffix=ext, dir=tmpDir, prefix=prefix)
    fileObject = os.fdopen(fd, "wb")
    return fileObject, filename

def joinMkdir(*args):
    """ join paths like os.path.join, do an mkdir, ignore all errors """
    path = os.path.join(*args)
    if not os.path.isdir(path):
        logging.debug("Creating dir %s" % path)
        os.makedirs(path)
    return path

def iterCsvRows(path, headers=None):
    " iterate over rows of csv file, uses the csv.reader, see below for homemade version "
    Rec = None
    for row in csv.reader(open(path, "rb")):
        if headers == None:
            headers = row
            headers = [re.sub("[^a-zA-Z]","_", h) for h in headers]
            Rec = collections.namedtuple("iterCsvRow", headers)
            continue
        fields = Rec(*row)
        yield fields

def iterTsvDir(inDir, ext=".tab.gz", prefix="", headers=None, format=None, fieldTypes=None, \
            noHeaderCount=None, encoding="utf8", fieldSep="\t", onlyFirst=False):
    " run iterTsvRows on all .tab or .tab.gz files in inDir "
    inMask = os.path.join(inDir, prefix+"*"+ext)
    inFnames = glob.glob(inMask)
    logging.debug("Found files %s" % inFnames)
    pm = ProgressMeter(len(inFnames))
    if len(inFnames)==0:
        raise Exception("No file matches %s" % inMask)

    for inFname in inFnames:
        if getsize(inFname)==0:
            logging.warn("%s has zero filesize, skipping" % inFname)
            continue
        for row in iterTsvRows(inFname, headers, format, fieldTypes, noHeaderCount, encoding, fieldSep):
            yield row
        pm.taskCompleted()
        if onlyFirst:
            break

def fastIterTsvRows(inFname):
    """
    simplistic version of iterTsvRows for higher speed.
    creates namedtuples from file and returns them AND THE LINE
    like iterTsvRows, but loads full file into memory.
    """
    if inFname.endswith(".gz"):
        openFunc = gzip.open
    else:
        openFunc = open
    headers = openFunc(inFname).readline().strip("\n").split("\t")
    Record = collections.namedtuple('tsvRec', headers)
    data = openFunc(inFname).read()
    #data = data.decode("utf8")
    lines = data.splitlines()
    for line in lines[1:]:
        yield Record(*line.split("\t")), line

class TsvReader():
    """ yields namedtuple objects from a tab-sep file. Was necessary as I needed a .seek function. """
    def __init__(self, ifh, encoding="utf8"):
        self.fieldNames = ifh.readline().lstrip("#").rstrip("\n").split("\t")
        self.Rec = collections.namedtuple('tsvRec', self.fieldNames)
        self.fieldCount = len(self.fieldNames)
        self.ifh = ifh
        self.encoding=encoding

    def nextRow(self):
        line = self.ifh.readline()
        if line=="":
            return None
        logging.log(5, "got line: %s" % line)
        cols = line.strip("\n").split("\t")
        if len(cols)!=self.fieldCount:
            raise Exception("headers not in sync with column count. headers: %s, column: %s" % (self.fieldNames, cols))

        cols = [c.decode("utf8") for c in cols]
        row = self.Rec(*cols)
        return row

    def seek(self, pos):
        self.ifh.seek(pos)

def iterTsvRows(inFile, headers=None, format=None, noHeaderCount=None, fieldTypes=None, fieldSep="\t", isGzip=False, skipLines=None, makeHeadersUnique=False, commentPrefix=None):
    """
        parses tab-sep file with headers as field names
        yields collection.namedtuples
        strips "#"-prefix from header line

        if file has no headers:

        a) needs to be called with
        noHeaderCount set to number of columns.
        headers will then be named col0, col1, col2, col3, etc...

        b) you can also set headers to a list of strings
        and supply header names in that way.

        c) set the "format" to one of: psl, this will also do type conversion

        fieldTypes can be a list of types.xxx objects that will be used for type
        conversion (e.g. types.IntType)

        - makeHeadersUnique will append _1, _2, etc to make duplicated headers unique.
        - skipLines can be used to skip x lines at the beginning of the file.
        - if encoding is None file will be read as byte strings.
        - commentPrefix specifies a character like "#" that markes lines to skip
    """

    if noHeaderCount:
        numbers = list(range(0, noHeaderCount))
        headers = ["col" + str(x) for x in numbers]

    if format=="psl":
        headers =      ["score", "misMatches", "repMatches", "nCount", "qNumInsert", "qBaseInsert", "tNumInsert", "tBaseInsert", "strand",    "qName",    "qSize", "qStart", "qEnd", "tName",    "tSize", "tStart", "tEnd", "blockCount", "blockSizes", "qStarts", "tStarts"]
        fieldTypes =   [IntType, IntType,      IntType,      IntType,  IntType,      IntType,       IntType,       IntType,      StringType,  StringType, IntType, IntType,  IntType,StringType, IntType, IntType,  IntType,IntType ,     StringType,   StringType,StringType]
    elif format=="bed12":
        headers =      ["chrom", "chromStart", "chromEnd", "name", "score", "strand", "thickStart", "thickEnd", "itemRgb",    "blockCount",    "blockSizes", "blockStarts"]
        fieldTypes =   [StringType, IntType,    IntType,    StringType,IntType,StringType,IntType,   IntType,      StringType,  IntType,        StringType, StringType]

    if isinstance(inFile, str):
        if inFile.endswith(".gz") or isGzip:
            #zf = gzip.open(inFile, 'rb')
            fh = gzip.open(inFile, 'rb')
            #reader = codecs.getreader(encoding)
            #fh = reader(zf)
        else:
            fh = open(inFile)
            #if encoding!=None:
                #fh = codecs.open(inFile, encoding=encoding)
            #else:
                #fh = open(inFile)
    else:
        fh = inFile

    if headers==None:
        line1 = fh.readline()
        line1 = line1.rstrip("\n").strip("#").rstrip("\t")
        headers = line1.split(fieldSep)
        headers = [h.strip() for h in headers]
        headers = [re.sub("[^a-zA-Z0-9_]","_", h) for h in headers]
        newHeaders = []
        for h in headers:
            if h[0].isdigit():
                h = "n"+h
            newHeaders.append(h)
        headers = newHeaders

    if makeHeadersUnique:
        newHeaders = []
        headerNum = defaultdict(int)
        for h in headers:
            headerNum[h]+=1
            if headerNum[h]!=1:
                h = h+"_"+str(headerNum[h])
            newHeaders.append(h)
        headers = newHeaders

    if skipLines:
        for i in range(0, skipLines):
            fh.readline()

    Record = collections.namedtuple('tsvRec', headers)
    for line in fh:
        if commentPrefix!=None and line.startswith(commentPrefix):
            continue
        line = line.strip("\n")
        fields = line.split(fieldSep)
        fields = [f for f in fields]
        if fieldTypes:
            fields = [f(x) for f, x in zip(fieldTypes, fields)]
        try:
            rec = Record(*fields)
        except Exception as msg:
            logging.error("Exception occured while parsing line, %s" % msg)
            logging.error("Filename %s" % fh.name)
            logging.error("Line was: %s" % line)
            logging.error("Does number of fields match headers?")
            logging.error("Headers are: %s" % headers)
            logging.error("Header field count is %d, line field count is %d" % (len(headers), len(fields)))
            raise Exception("wrong field count in line %s" % line)
        # convert fields to correct data type
        yield rec

def iterTsvGroups(fileObject, **kwargs):
    """
    iterate over a tab sep file, convert lines to namedtuples (records), group lines by some field.

    file needs to be sorted on this field!
    parameters:
        groupFieldNumber: number, the index (int) of the field to group on
        useChar: number, only use the first X chars of the groupField
        groupFieldSep: a char, uses only the part before this character in the groupField

    return:
        (groupId, list of namedtuples)
    """
    groupFieldNumber = kwargs.get("groupFieldNumber", 0)
    useChars = kwargs.get("useChars", None)
    groupFieldSep = kwargs.get("groupFieldSep", None)
    if "groupFieldNumber" in kwargs:
        del kwargs["groupFieldNumber"]
    if "groupFieldSep" in kwargs:
        del kwargs["groupFieldSep"]
    if useChars:
        del kwargs["useChars"]
    assert(groupFieldNumber!=None)

    lastId = None
    group = []
    id = None
    for rec in iterTsvRows(fileObject, **kwargs):
        id = rec[groupFieldNumber]
        if useChars:
            id = id[:useChars]
        if groupFieldSep:
            id = id.split(groupFieldSep)[0]
        if lastId==None:
            lastId = id
        if lastId==id:
            group.append(rec)
        else:
            yield lastId, group
            group = [rec]
            lastId = id
    if id!=None:
        yield id, group

def iterTsvJoin(files, **kwargs):
    r"""
    iterate over two sorted tab sep files, join lines by some field and yield as namedtuples
    files need to be sorted on the field!

    parameters:
        groupFieldNumber: number, the index (int) of the field to group on
        useChar: number, only use the first X chars of the groupField

    return:
        yield tuples (groupId, [file1Recs, file2Recs])
    >>> f1 = StringIO.StringIO("id\ttext\n1\tyes\n2\tunpaired middle\n3\tno\n5\tnothing either\n")
    >>> f2 = StringIO.StringIO("id\ttext\n0\tnothing\n1\tvalid\n3\tnot valid\n")
    >>> files = [f1, f2]
    >>> list(iterTsvJoin(files, groupFieldNumber=0))
    [(1, [[tsvRec(id='1', text='yes')], [tsvRec(id='1', text='valid')]]), (3, [[tsvRec(id='3', text='no')], [tsvRec(id='3', text='not valid')]])]
    """
    assert(len(files)==2)
    f1, f2 = files
    iter1 = iterTsvGroups(f1, **kwargs)
    iter2 = iterTsvGroups(f2, **kwargs)
    groupId1, recs1 = next(iter1)
    groupId2, recs2 = next(iter2)
    while True:
        groupId1, groupId2 = int(groupId1), int(groupId2)
        if groupId1 < groupId2:
            groupId1, recs1 = next(iter1)
        elif groupId1 > groupId2:
            groupId2, recs2 = next(iter2)
        else:
            yield groupId1, [recs1, recs2]
            groupId1, recs1 = next(iter1)
            groupId2, recs2 = next(iter2)

def runCommand(cmd, ignoreErrors=False, verbose=False):
    """ run command in shell, exit if not successful """
    #if type(cmd)==types.ListType:
        #cmd = " ".join(cmd)
    msg = "Running shell command: %s" % cmd
    logging.debug(msg)
    if verbose:
        logging.info(msg)

    if type(cmd)==bytes:
        ret = os.system(cmd)
    elif type(cmd)==list:
        ret = subprocess.call(cmd)
        cmd = " ".join(cmd) # for debug output
    else:
        assert(False) # has to be called with string or list

    if ret!=0:
        if ignoreErrors:
            logging.info("Could not run command %s, retcode %s" % (cmd, str(ret)))
            return None
        else:
            raise Exception("Could not run command (Exitcode %d): %s" % (ret, cmd))
    return ret

def makedirs(path, quiet=False):
    try:
        os.makedirs(path)
    except:
        if not quiet:
            raise

def appendTsvNamedtuple(filename, row):
    " append a namedtuple to a file. Write headers if file does not exist "
    if not os.path.isfile(filename):
       outFh = open(filename, "w")
       headers = row._fields
       outFh.write("\t".join(headers)+"\n")
    else:
       outFh = open(filename, "a")
    outFh.write("\t".join(row)+"\n")

def appendTsvDict(filename, inDict, headers):
    " append a dict to a file in the order of headers"
    values = []
    if headers==None:
        headers = list(inDict.keys())

    for head in headers:
        values.append(inDict.get(head, ""))

    logging.log(5, "order of headers is: %s" % headers)

    if not os.path.isfile(filename):
       outFh = codecs.open(filename, "w", encoding="utf8")
       outFh.write("\t".join(headers)+"\n")
    else:
       outFh = codecs.open(filename, "a", encoding="utf8")
    logging.log(5, "values are: %s" % values)
    values = [x if x !=None else "" for x in values]
    outFh.write("\t".join(values)+"\n")

def appendTsvOrderedDict(filename, orderedDict):
    appendTsvDict(filename, orderedDict, None)

class ProgressMeter:
    """ prints a message "x%" every stepCount/taskCount calls of taskCompleted()
    """
    def __init__(self, taskCount, stepCount=20, quiet=False):
        self.taskCount=taskCount
        self.stepCount=stepCount
        self.tasksPerMsg = taskCount/stepCount
        self.i=0
        self.quiet = quiet
        #print "".join(9*["."])

    def taskCompleted(self, count=1):
        if self.quiet and self.taskCount<=5:
            return
        #logging.debug("task completed called, i=%d, tasksPerMsg=%d" % (self.i, self.tasksPerMsg))
        if self.tasksPerMsg!=0 and self.i % self.tasksPerMsg == 0:
            donePercent = (self.i*100) / self.taskCount
            #print "".join(5*[chr(8)]),
            sys.stderr.write("%.2d%% " % donePercent)
            sys.stderr.flush()
        self.i += count
        if self.i==self.taskCount:
            print("")

def test():
    pm = ProgressMeter(2000)
    for i in range(0,2000):
        pm.taskCompleted()

def parseConfig(f):
    " parse a name=value file from file-like object f and return as dict"
    if isinstance(f, str):
        logging.debug("parsing config file %s" % f)
        f = open(os.path.expanduser(f))
    result = {}
    for line in f:
        if line.startswith("#"):
            continue
        line = line.strip()
        if "=" in line:
            key, val = line.split("=")
            result[key]=val
    return result

def retryHttpRequest(url, params=None, repeatCount=15, delaySecs=120, userAgent=None, onlyHead=False):
    """ wrap urlopen in try...except clause and repeat
    #>>> retryHttpHeadRequest("http://www.test.com", repeatCount=1, delaySecs=1)
    """

    class HeadRequest(urllib.request.Request):
        def get_method(self):
            return 'HEAD'

    def handleEx(ex, count):
        logging.info("Got Exception %s, %s on urlopen of %s, %s. Waiting %d seconds before retry..." % \
            (type(ex), str(ex), url, params, delaySecs))
        time.sleep(delaySecs)
        count = count - 1
        return count

    socket.setdefaulttimeout(20)
    count = repeatCount
    while count>0:
        try:
            logging.log(5, "Getting URL %s, params %s" % (url, params))
            if onlyHead:
                req = HeadRequest(url, params)
            else:
                req = urllib.request.Request(url, params)
            if userAgent != None:
                req.add_header('User-Agent', userAgent)
            opener = urllib.request.build_opener()
            ret = opener.open(req, timeout=20)
            #ret = urllib2.urlopen(url, params, 20)
        except urllib.error.HTTPError as ex:
            count = handleEx(ex, count)
        except http.client.HTTPException as ex:
            count = handleEx(ex, count)
        except urllib.error.URLError as ex:
            count = handleEx(ex, count)
        except socket.timeout as ex:
            count = handleEx(ex, count)
        except socket.error as ex:
            count = handleEx(ex, count)
        else:
            return ret

    logging.debug("Got repeatedexceptions on urlopen, returning None")
    return None

def retryHttpHeadRequest(url, repeatCount=15, delaySecs=120, userAgent = None):
    response = retryHttpRequest(url, repeatCount=repeatCount, delaySecs=delaySecs, \
        userAgent=userAgent, onlyHead=True)
    return response

def sendEmail(address, subject, text):
    text = text.replace("'","")
    subject = subject.replace("'","")
    cmd = "echo '%s' | mail -s '%s' %s" % (text, subject, address)
    logging.info("Email command %s" % cmd)
    os.system(cmd)

def getAppDir():
    """ get base directory of the package, usually the dir where the scripts
    are. A frozen package is a binary file that includes python and the source code
    into an executable.
    """
    logging.debug("getAppDir: executable is %s" % sys.executable)
    if getattr(sys, 'frozen', False):
        # frozen
        appDir = dirname(sys.executable)
    else:
        # unfrozen
        appDir = abspath(dirname(dirname(realpath(__file__))))
    return appDir

if __name__=="__main__":
    #test()
    import doctest
    doctest.testmod()



# package to download and parse xml files from pubmed/medline
# + some functions to query eutils

import logging, urllib.request, urllib.error, urllib.parse, re, time, urllib.request, urllib.parse, urllib.error, traceback, http.client
from . import pubConf, maxXml, pubStore, maxCommon
import socket

from xml.etree.ElementTree import ParseError
from xml.etree.cElementTree import ParseError as ParseError2
from collections import OrderedDict

class PubmedError(Exception):
    def __init__(self, longMsg, logMsg, detailMsg=None):
        self.longMsg = longMsg
        self.logMsg = logMsg
        self.detailMsg = detailMsg

    def __str__(self):
        return repr(self.longMsg+"/"+self.logMsg)

def getMedlineDate(medlineData, dateField):
    year = medlineData.getTextFirst("{}/Year".format(dateField))
    if (year is None) or (len(year) == 0):
        return ""
    else:
        return "{}-{}-{}".format(year,
                                 medlineData.getTextFirst("{}/Month".format(dateField)),
                                 medlineData.getTextFirst("{}/Day".format(dateField)))

def getMedlineText(roots):
    """Obtain text from title and abstracts nodes. Keep inline markup. For
    AbstractText tags with NlmCategory attributes, the attribute value is
    added to the abstract text, which makes for nice reading."""
    # FIXME: can this be merged with treeToAsciiText?
    textParts = []
    for root in roots:
        text = root.getText()
        for child in root.getXmlAll('*'):
            text += getMedlineText([child])
        text += root.getTextTail()
        textParts.append(text)
    return "".join(textParts).strip()

def parseMedline(xmlParser):
    """
    fill article data dict with pubmed xml data

    >>> xml = PubmedTestDoc()
    >>> data = parseMedline(maxXml.XmlParser(string=xml))
    >>> del data["time"]
    >>> repr(data)
     "OrderedDict([('articleId', ''), ('externalId', 'PMID20430833'), ('source', ''), ('origFile', ''), ('journal', 'Brain : a journal of neurology'), ('printIssn', '0006-8950'), ('eIssn', '0006-8950'), ('journalUniqueId', '0372537'), ('year', '2010'), ('articleType', 'research-article'), ('articleSection', ''), ('authors', u'Willemsen, Mich\\\\xe9l A; Verbeek, Marcel M'), ('authorEmails', ''), ('authorAffiliations', 'Radboud University Nijmegen Medical Centre, Donders Institute for Brain, Cognition and Behaviour, Department of Paediatric Neurology (820 IKNC), PO Box 9101, 6500 HB Nijmegen, The Netherlands. m.willemsen@cukz.umcn.nl'), ('keywords', 'Age of Onset/Useless Research'), ('title', 'Tyrosine hydroxylase deficiency: a treatable disorder of brain catecholamine biosynthesis.'), ('abstract', 'An infantile onset, progressive, hypokinetic-rigid syndrome with dystonia (type A), and a complex encephalopathy with neonatal onset (type B). Decreased cerebrospinal fluid concentrations of homovanillic acid and c.698G>A and c.707T>C mutations. Carriership of at least one promotor mutation, however, apparently predicts type A tyrosine hydroxylase deficiency. Most patients with tyrosine hydroxylase deficiency can be successfully treated with l-dopa.'), ('vol', '133'), ('issue', 'Pt 6'), ('page', '1810-22'), ('pmid', '20430833'), ('pmcId', ''), ('doi', ''), ('fulltextUrl', 'https://www.ncbi.nlm.nih.gov/pubmed/20430833')])"

    """
    data = pubStore.createEmptyArticleDict()
    #medlineData           = xmlParser.getXmlFirst("MedlineCitation")
    medlineData           = xmlParser
    data["pmid"]          = medlineData.getTextFirst("PMID")
    el = medlineData.getElFirst("PMID", None)
    data["pmidVersion"]   = el.attrib.get("Version", "") if el is not None else ""
    data["externalId"]    = "PMID"+data["pmid"]
    data["fulltextUrl"]   = "https://www.ncbi.nlm.nih.gov/pubmed/%s" % data["pmid"]
    logging.log(5, "PMID %s" % data["pmid"])
    data["medlineCreatedDate"] = getMedlineDate(medlineData, "DateCreated")
    data["medlineCompletedDate"] = getMedlineDate(medlineData, "DateCompleted")
    data["medlineRevisedDate"] = getMedlineDate(medlineData, "DateRevised")
    otherIds         = medlineData.getTextAll("OtherID", reqAttrDict={"Source":"NLM"})
    pmcIds = [i for i in otherIds if i.startswith("PMC")]
    if len(pmcIds) > 0:
        data["pmcId"] = pmcIds[0].split()[0].replace("PMC","")

    artTree               = medlineData.getXmlFirst("Article")

    data["title"]         = getMedlineText(artTree.getXmlAll("ArticleTitle"))

    # handle structured abstracts
    data["abstract"]      = getMedlineText(artTree.getXmlAll("Abstract/AbstractText"))

    if data["abstract"]=="":
        data["abstract"]      = getMedlineText(artTree.getXmlAll("OtherAbstract/AbstractText"))

    data["authorAffiliations"]   = artTree.getTextFirst("Affiliation", default="")
    data["doi"]           = artTree.getTextFirst("ELocationID", default="", reqAttrDict={"EIdType":"doi"})
    data["lang"]   = artTree.getTextFirst("Language", default="")

    data["journalUniqueId"] = medlineData.getTextFirst("MedlineJournalInfo/NlmUniqueID")
    linkingIssn = medlineData.getTextFirst("MedlineJournalInfo/ISSNLinking", default="")

    journalTree = artTree.getXmlFirst("Journal")
    data["eIssn"]       = journalTree.getTextFirst("ISSN", reqAttrDict={"IssnType": 'Electronic'}, default="")
    data["printIssn"]   = journalTree.getTextFirst("ISSN", reqAttrDict={"IssnType": 'Print'}, default="")
    # keep the link ISSN when we have space, e.g. PNAS is not storing the print ISSN anymore, only as link Issn
    if data["printIssn"]=="" and linkingIssn!="":
        data["printIssn"]   = linkingIssn
    if data["eIssn"]=="" and linkingIssn!="":
        data["eIssn"]   = linkingIssn

    data["vol"]         = journalTree.getTextFirst("JournalIssue/Volume", default="")
    data["issue"]       = journalTree.getTextFirst("JournalIssue/Issue", default="")
    data["year"]        = journalTree.getTextFirst("JournalIssue/PubDate/Year", default="")
    if data["year"]=="":
        year = journalTree.getTextFirst("JournalIssue/PubDate/MedlineDate", default="").split()[0]
        if not year.isdigit():
            year = ""
        data["year"] = year
    data["journal"]     = journalTree.getTextFirst("Title", default="")
    data["page"]        = artTree.getTextFirst("Pagination/MedlinePgn", default="")

    authorList  = artTree.getXmlFirst("AuthorList")
    lastNames   = []
    initialList = []
    if authorList!=None:
        authorTrees = authorList.getXmlAll("Author")
        for authorTree in authorTrees:
            lastName = authorTree.getTextFirst("LastName", default="")
            if lastName=="":
                lastName = authorTree.getTextFirst("CollectiveName", default="")
            lastNames.append(lastName)

            initials = authorTree.getTextFirst("ForeName", default="")
            if initials=="":
                initials = authorTree.getTextFirst("Initials", default="")
            initialList.append(initials)

    authors = [lastNames[i]+", "+initialList[i] for i in range(0, min(len(lastNames), len(initialList)))]
    data["authors"]="; ".join(authors)

    articleTypeList = set(artTree.getTextAll("PublicationTypeList/PublicationType"))
    articleTypesString  = ",".join(articleTypeList)

    articleType="research-article"

    noResearchArticleTags = ["Bibliography", "Biography",
        "Case Reports", "Webcasts",
        "Dictionary", "Directory",
        "Editorial", "Festschrift",
        "Patient Education Handout", "Periodical Index",
        "Portraits", "Published Erratum", "Scientific Integrity Review"
        "Congresses"]

    if "Review" in articleTypeList:
       articleType = "review"
    elif "Letter" in articleTypeList:
       articleType = "research-article"
    else:
        for noResearchArticleTag in noResearchArticleTags:
            if noResearchArticleTag in articleTypeList:
                articleType = "other"
                break

    data["articleType"]        = articleType
    #data["pubmedArticleTypes"] = articleTypesString

    logging.log(5, "pubmedArticleTypes %s, articleType %s" % (articleTypesString, articleType))

    meshDescriptors = []
    meshHeadingList       = medlineData.getXmlFirst("MeshHeadingList", default="")
    if meshHeadingList:
        #for meshHeadingDescriptor in meshHeadingList.getTextAll("MeshHeading/DescriptorName", reqAttrDict={"MajorTopicYN":"Y"}):
        for meshHeadingDescriptor in meshHeadingList.getTextAll("MeshHeading/DescriptorName"):
            meshDescriptors.append(meshHeadingDescriptor.strip())

    data["keywords"] = "/".join(meshDescriptors)

    # remove these annoying linebreaks!
    filtData = {}
    for key, val in data.items():
        filtData[key] = val.replace('\u2028', ' ')
    return filtData

def parsePubmedFields(xmlEl, dataDict):
    """ parse special pubmed (not medline) fields, like doi, pmc, etc """
    dataDict["doi"]           = xmlEl.getTextFirst("ArticleIdList/ArticleId", reqAttrDict = {"IdType" : 'doi'}, default="")
    dataDict["pii"]           = xmlEl.getTextFirst("ArticleIdList/ArticleId", reqAttrDict = {"IdType" : 'pii'}, default="")
    dataDict["mid"]           = xmlEl.getTextFirst("ArticleIdList/ArticleId", reqAttrDict = {"IdType" : 'mid'}, default="")
    dataDict["pmcId"]         = xmlEl.getTextFirst("ArticleIdList/ArticleId", reqAttrDict = {"IdType" : 'pmc'}, default="").replace("PMC", "")
    verEl = xmlEl.getXmlFirst("PMID")
    return dataDict

def parsePubmedMedlineIter(xml):
    """
    Parse pubmed xml format and yield as dictionary, see parseMedline
    records come either from Pubmed as a <PubmedArticleSet><PubmedArticle><MedlineCitation>...
    or from Medline as <MedlineCitationSet><MedlineCitation>...</MedlineCitation>
    """
    #if fromMedline:
        #recordTag = "MedlineCitation"
        #closeTag = "</MedlineArticleSet>"
        #openTag = "<MedlineArticleSet"
    #else:
    recordTag = "PubmedArticle"
    #recordTag = "PubmedArticle/MedlineCitation"
    closeTag = b"</PubmedArticleSet>"
    openTag = b"<PubmedArticleSet>"
    # NCBI eutils sometimes "forgets" the opening/closing tags
    if xml.strip()=="":
        logging.error("Got empty XML file from NCBI")
        raise PubmedError("Got empty XML from NCBI", "pubmedEmptyXml")

    if not openTag in xml:
        logging.warn("Addding opening tag")
        xml = openTag + "\n" + xml

    if not closeTag in xml:
        logging.warn("Addding closing tag")
        xml = xml+"\n"+closeTag

    logging.debug("Parsing pubmed file")
    try:
        topEl = maxXml.XmlParser(string=xml)
    except ParseError:
        logging.debug("Error on parsing this XML: %s" % xml)
        raise
    except ParseError2:
        logging.debug("Error on parsing this XML: %s" % xml)
        raise

    for artEl in topEl.getXmlAll(recordTag):
        medlineCitEl = artEl.getXmlFirst("MedlineCitation")
        dataDict = parseMedline(medlineCitEl)

        pubmedCitEl = artEl.getXmlFirst("PubmedData")
        dataDict = parsePubmedFields(pubmedCitEl, dataDict)
        yield dataDict

def ncbiEFetchGenerator(ids, dbName="pubmed", tool="pubtools", email=pubConf.email, debug=False):
    """
    retrieve records in xml format from ncbi, and yield as dictionaries
    >> print len(list(ncbiEFetchGenerator(["9322214"], debug=True)))
    1
    """
    idsLeft=list(set(ids))
    retmax = 500

    downloadCount=0
    while len(idsLeft)!=0:
        retStart = downloadCount
        downloadIds = idsLeft[:min(retmax, len(idsLeft))]
        logging.debug("Getting data on %d PMIDs from NCBI, %d PMIDs left to download" % (len(downloadIds), len(idsLeft)))
        url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=%s&tool=%s&email=%s&retmode=xml&rettype=medline&id=%s" % (dbName, tool, email, ",".join(downloadIds))
        logging.debug("Getting %s" % url)

        tryCount = 0
        xml = None
        while tryCount < 10 and xml == None:
            try:
                tryCount += 1
                xml = urllib.request.urlopen(url.strip()).read()
                if xml == None:
                    raise Exception("Could not connect to Eutils")

                idsLeft= list(set(idsLeft).difference(downloadIds))
                for pubmedData in parsePubmedMedlineIter(xml):
                    downloadCount+=1
                    yield pubmedData
            #except urllib2.HTTPError:
                ##I sometimes see "HTTP Error 502: Bad Gateway"
                #logging.info("HTTP Error on eutils, pausing for 120 secs")
                #time.sleep(120)
            except socket.error:
                logging.info("Socket error on eutils, pausing for 120 secs")
                time.sleep(120)
            except urllib.error.URLError: # this should handle HTTPError, too
                logging.info("HTTP Error on eutils, pausing for 120 secs")
                time.sleep(120)
            except socket.timeout:
                logging.info("Socket timeout on eutils, pausing for 120 secs")
                time.sleep(120)
            except http.client.BadStatusLine:
                logging.info("Bad status line on eutils, pausing 120 secs")
                time.sleep(120)
            except http.client.IncompleteRead: # this happened only once in three years
                logging.info("IncompleteRead Error on eutils, pausing for 120 secs")
                time.sleep(120)
            except ParseError:
                logging.info("XML ParseError on eUtils, pausing 120 secs")
                time.sleep(120)

def ncbiESearch(query, dbName="pubmed", tool="", email="maximilianh@gmail.com", debug=False, delaySecs=0):
    """ retrieve pmids for query, returns list of ids
    >> len(set(ncbiESearch("human genome", debug=True)))
    66009
    """
    if debug:
        logging.getLogger().setLevel(5)
        logging.debug("debug mode activated")

    retmax=100000
    addString=""
    query = urllib.parse.quote(query)

    idsLeft = None
    allPmids = []

    while idsLeft>0 or idsLeft==None:
        logging.debug( "PMIDs left %s, PMIDs downloaded %d" % (str(idsLeft), len(allPmids)))
        url = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?db=%s&tool=retrPubmed&tool=%s&email=%s&term=%s&retstart=%d&retmax=%d%s' % (dbName, tool, email, query, len(allPmids), retmax, addString)
        req = urllib.request.Request(url)
        html = urllib.request.urlopen(req)
        logging.debug("Getting "+url+"\n")

        pmids = []
        for line in html:
            if line.find("<Count>")!=-1:
                if idsLeft==None: # only use the first "count" line
                    fs = re.split("<.{0,1}Count>", line)
                    idsLeft = int(fs[1])
            if line.find("<Id>")!=-1:
                pmid = line.strip().replace("<Id>","").replace("</Id>", "")
                pmids.append(pmid)
        idsLeft -= len(pmids)
        allPmids.extend(pmids)
        logging.info("Sleeping %d seconds" % delaySecs)
        time.sleep(delaySecs)

    return allPmids

def PubmedTestDoc():
    return '''
        <MedlineCitation Owner="NLM" Status="MEDLINE">
        <PMID>20430833</PMID>
        <DateCreated>
            <Year>2010</Year>
            <Month>05</Month>
            <Day>31</Day>
        </DateCreated>
        <DateCompleted>
            <Year>2010</Year>
            <Month>06</Month>
            <Day>22</Day>
        </DateCompleted>
        <Article PubModel="Print-Electronic">
            <Journal>
                <ISSN IssnType="Print">1460-1234</ISSN>
                <ISSN IssnType="Electronic">1460-2156</ISSN>
                <JournalIssue CitedMedium="Internet">
                    <Volume>133</Volume>
                    <Issue>Pt 6</Issue>
                    <PubDate>
                        <Year>2010</Year>
                        <Month>Jun</Month>
                    </PubDate>
                </JournalIssue>
                <Title>Brain : a journal of neurology</Title>
                <ISOAbbreviation>Brain</ISOAbbreviation>
            </Journal>
            <ArticleTitle>Tyrosine hydroxylase deficiency: a treatable disorder of brain catecholamine biosynthesis.</ArticleTitle>
            <Pagination>
                <MedlinePgn>1810-22</MedlinePgn>
            </Pagination>
            <Abstract>
                <AbstractText>An infantile onset, progressive, hypokinetic-rigid syndrome with dystonia (type A), and a complex encephalopathy with neonatal onset (type B). Decreased cerebrospinal fluid concentrations of homovanillic acid and c.698G&gt;A and c.707T&gt;C mutations. Carriership of at least one promotor mutation, however, apparently predicts type A tyrosine hydroxylase deficiency. Most patients with tyrosine hydroxylase deficiency can be successfully treated with l-dopa.</AbstractText>
            </Abstract>
            <Affiliation>Radboud University Nijmegen Medical Centre, Donders Institute for Brain, Cognition and Behaviour, Department of Paediatric Neurology (820 IKNC), PO Box 9101, 6500 HB Nijmegen, The Netherlands. m.willemsen@cukz.umcn.nl</Affiliation>
            <AuthorList CompleteYN="Y">
                <Author ValidYN="Y">
                    <LastName>Willemsen</LastName>
                    <ForeName>Mich&#233;l A</ForeName>
                    <Initials>MA</Initials>
                </Author>
                <Author ValidYN="Y">
                    <LastName>Verbeek</LastName>
                    <ForeName>Marcel M</ForeName>
                    <Initials>MM</Initials>
                </Author>
            </AuthorList>
            <Language>eng</Language>
            <PublicationTypeList>
                <PublicationType>Journal Article</PublicationType>
                <PublicationType>Research Support, Non-U.S. Gov't</PublicationType>
            </PublicationTypeList>
            <ArticleDate DateType="Electronic">
                <Year>2010</Year>
                <Month>04</Month>
                <Day>29</Day>
            </ArticleDate>
        </Article>
        <MedlineJournalInfo>
            <Country>England</Country>
            <MedlineTA>Brain</MedlineTA>
            <NlmUniqueID>0372537</NlmUniqueID>
            <ISSNLinking>0006-8950</ISSNLinking>
        </MedlineJournalInfo>
        <ChemicalList>
            <Chemical>
                <RegistryNumber>0</RegistryNumber>
                <NameOfSubstance>Catecholamines</NameOfSubstance>
            </Chemical>
        </ChemicalList>
        <CitationSubset>AIM</CitationSubset>
        <CitationSubset>IM</CitationSubset>
        <MeshHeadingList>
            <MeshHeading>
                <DescriptorName MajorTopicYN="N">Age of Onset</DescriptorName>
            </MeshHeading>
            <MeshHeading>
                <DescriptorName MajorTopicYN="Y">Useless Research</DescriptorName>
            </MeshHeading>
        </MeshHeadingList>
        </MedlineCitation>
        '''

def getOnePmid(pmid):
    " fill out article data dict via eutils service for one single pmid "
    #logging.debug("Getting article meta data from NCBI via Eutils for PMID %s" % str(pmid))
    articleDataList = list(ncbiEFetchGenerator([pmid]))
    if len(articleDataList)==0:
        logging.info("No data from Pubmed for PMID %s" % str(pmid))
        return None
    else:
        pubMeta = articleDataList[0]
        logging.log(5, pubMeta)
        return pubMeta

def stripTag(line):
    remHtml = re.compile("<(.|\n)*?>")
    line = re.sub(remHtml, "", line)
    line = line.strip()
    return line

def getOutlinks(pmid, preferPmc=False):
    """ use NCBI eutils to get outlinks for a pmid as a dict provider -> url  """
    logging.debug("%s: Getting outlink from pubmed" % (pmid))
    url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi?dbfrom=pubmed&id=%s&retmode=llinks&cmd=llinks" % pmid
    #logging.debug("getting %s" % url)
    #try:
    #except
    #logging.info(traceback.format_exc())
    #logging.info("Exception when downloading")
    #return None
    userAgent = 'User-Agent: Mozilla (max@soe.ucsc.edu, http://text.soe.ucsc.edu)'

    #html = urllib2.urlopen(req)
    html = maxCommon.retryHttpRequest(url, userAgent=userAgent)

    if html==None:
        return None

    outlinks = OrderedDict()
    provider = False
    fullText = False
    isPmc = False

    for line in html:
        if line.find("<ObjUrl>") != -1:
            url=""
            fullText=False
            origPublisher=False
        if line.find("Attribute") != -1:
            attribute=stripTag(line)
            if attribute.lower()=="full-text online" or attribute=="full-text pdf":
                fullText=True
        if line.find("<NameAbbr>") != -1 and fullText and origPublisher:
            db = stripTag(line)
            outlinks[db]=url
        if line.find("publishers/providers")!=-1:
            origPublisher=True
        if line.find("<Provider>") != -1:
            provider=True
        if line.find("</Provider>") != -1:
            provider=False
        if line.find("<DbFrom>") != -1:
            db = stripTag(line)
        if line.find("<Url>") != -1 and not provider:
            url = line
            url = stripTag(url).replace("&amp;", "&") # XX strange!
            url = stripTag(url).replace("&lt;", "<")
            url = stripTag(url).replace("&gt;", ">")
            if "www.ncbi.nlm.nih.gov" in url and "/pmc/" in url and preferPmc:
                # override all other links
                outlinks.clear()
                outlinks["pmc"] = url

    logging.debug("%s: Found outlinks %s" % (pmid, str(outlinks)))
    return outlinks

if __name__=="__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
import logging, re, urllib.request, urllib.parse, urllib.error
import xml.etree.ElementTree as etree

class XmlParser(object):
    """ class to represent an xml tree (using ElementTree)
        Functions Accept PATH which is a /a/b/c style xpath-like expression to refer to elements
        PATH is not a complete XPATH implementation

        getText... functions return just a string
        getXml... functions return an XmlParser-object
        ...First  functions get only the first instance
        ...All    functions return an iterator

    >>> xp = XmlParser(string="<fruit><apple size='big'>boskoop</apple><apple size='small'>granny smith</apple><pear>mypear</pear></fruit>")
    >>> xp.getTextFirst("pineapple", default="NothingAtAll")
    'NothingAtAll'
    >>> xp.getTextFirst("apple")
    'boskoop'
    >>> list(xp.getTextAll("apple"))
    ['boskoop', 'granny smith']
    >>> list(xp.getTextAll("apple", reqAttrDict={'size':'big'}))
    ['boskoop']

    """
    def __init__(self, string=None, url=None, root=None, removeNamespaces=False):
        self.root=None
        if string!=None:
            self.fromString(string, removeNamespaces)
        elif url!=None:
            self.fromUrl(url, removeNamespaces)
        elif root!=None:
            self.root=root

    def getAttr(self, name):
        return self.root.attrib.get(name, None)

    def getText(self):
        if self.root.text==None:
            return ""
        else:
            return self.root.text

    def getTextTail(self):
        if self.root.tail==None:
            return ""
        else:
            return self.root.tail

    def fromString(self, string, removeNamespaces=False):
        if string=="":
            return None
        root = etree.fromstring(string)
        if removeNamespaces:
            logging.debug("Stripping all namespaces")
            strip_namespace_inplace(root)
        self.root = root

    def fromUrl(self, url, removeNamespaces=False, stopWords=[]):
        logging.debug("Retrieving %s" % url)
        text = urllib.request.urlopen(url).read()
        self.fromString(text, removeNamespaces=removeNamespaces)
        #for w in stopWords:
            #if w in text:
                #return None

    def _removeNamespaces(self):
        """ removes all namespaces from elementtree IN PLACE """
        root = self.root
        for el in root.getiterator():
            if el.tag[0] == '{':
                el.tag = el.tag.split('}', 1)[1]

    def _hasAttribs(self, el, reqAttrDict):
        for attr, value in reqAttrDict.items():
            if el.attrib.get(attr, None)!=value:
                return False
        return True

    def getTextFirst(self, path, reqAttrDict=None, default=None):
        """ return text between elements given path
            reqAttrDict is in the format attrName -> value
        """
        xml = self.getElFirst(path, reqAttrDict)
        if xml != None and xml.text!=None:
            return xml.text
        else:
            return default

    def getTextAll(self, path, reqAttrDict=None):
        for el in self.getElAll(path, reqAttrDict):
            yield el.text

    def getElFirst(self, path, reqAttrDict):
        found = False
        for el in self.getElAll(path, reqAttrDict):
            found = True
            return el

    def getElAll(self, path, reqAttrDict):
        found = False
        elIter = self.root.findall(path)
        for el in elIter:
            if reqAttrDict == None or self._hasAttribs(el, reqAttrDict):
                found = True
                yield el

    def getXmlFirst(self, path, reqAttrDict=None, default=None):
        el = self.getElFirst(path, reqAttrDict)
        if el==None:
            return default
        else:
            return XmlParser(root=el)

    def getXmlAll(self, path, reqAttrDict=None):
        for el in self.getElAll(path, reqAttrDict):
            yield XmlParser(root=el)

    def __repr__(self):
        return etree.tostring(self.root, encoding='unicode')

def strip_namespace_inplace(etree, namespace=None,remove_from_attr=True):
    """ Takes a parsed ET structure and does an in-place removal of all namespaces,
        or removes a specific namespacem (by its URL).

        Can make node searches simpler in structures with unpredictable namespaces
        and in content given to be non-mixed.

        By default does so for node names as well as attribute names.
        (doesn't remove the namespace definitions, but apparently
         ElementTree serialization omits any that are unused)

        Note that for attributes that are unique only because of namespace,
        this may attributes to be overwritten.
        For example: <e p:at="bar" at="quu">   would become: <e at="bar">

        I don't think I've seen any XML where this matters, though.
    """
    if namespace==None: # all namespaces
        for elem in etree.getiterator():
            tagname = elem.tag
            if not isinstance(elem.tag, str):
                continue
            if tagname[0]=='{':
                elem.tag = tagname[ tagname.index('}',1)+1:]

            if remove_from_attr:
                to_delete=[]
                to_set={}
                for attr_name in elem.attrib:
                    if attr_name[0]=='{':
                        old_val = elem.attrib[attr_name]
                        to_delete.append(attr_name)
                        attr_name = attr_name[attr_name.index('}',1)+1:]
                        to_set[attr_name] = old_val
                for key in to_delete:
                    elem.attrib.pop(key)
                elem.attrib.update(to_set)

    else: # asked to remove specific namespace.
        ns = '{%s}' % namespace
        nsl = len(ns)
        for elem in etree.getiterator():
            if elem.tag.startswith(ns):
                elem.tag = elem.tag[nsl:]

            if remove_from_attr:
                to_delete=[]
                to_set={}
                for attr_name in elem.attrib:
                    if attr_name.startswith(ns):
                        old_val = elem.attrib[attr_name]
                        to_delete.append(attr_name)
                        attr_name = attr_name[nsl:]
                        to_set[attr_name] = old_val
                for key in to_delete:
                    elem.attrib.pop(key)
                elem.attrib.update(to_set)

# -----
if __name__ == "__main__":
    import doctest
    doctest.testmod()



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

setup(
    name='pubMunch3',
    version='1.0.1',
    description='A library and tool for downloading pubmed articles',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    install_requires=[
        'beautifulsoup4==4.6',
        'unidecode==1.0.22',
        'incapsula-cracker-py3==0.1.8.1',
        'html5lib==1.0.1'
    ],

    entry_points={
        'console_scripts': [
            'download_pmid=pubMunch:download_pmid_program',
        ],
    },
)
